"""Curiosity module define special construct with curio framework."""

from typing import List

from .definition import FAILURE, SUCCESS, AsyncInnerFunction, CallableFunction, node_metadata
from .utils import amap

__all__ = ['parallele']


try:
    from curio import gather, spawn

    def parallele(children: List[CallableFunction], succes_threshold: int = -1) -> AsyncInnerFunction:
        """Return an awaitable function which run children in parallele.

        `succes_threshold` parameter generalize traditional sequence/fallback,
        and must be in [0, len(children)], default value is len(children)

        if #success = succes_threshold, return a success

        if #failure = len(children) - succes_threshold, return a failure

        Args:
            children (List[CallableFunction]): list of Awaitable
            succes_threshold (int): succes threshold value, default len(children)

        Returns:
            (AsyncInnerFunction): an awaitable function.

        """
        succes_threshold = succes_threshold if succes_threshold != -1 else len(children)
        if not (0 <= succes_threshold <= len(children)):
            raise AssertionError('succes_threshold')

        @node_metadata(properties=['succes_threshold'])
        async def _parallele():

            results = await gather([task async for task in amap(spawn, children)])

            success = len(list(filter(bool, results)))

            return SUCCESS if success >= succes_threshold else FAILURE

        return _parallele


except Exception:  # pragma: no cover
    # default to a simple sequence
    from .control import sequence as parallele



"""Common definition.

CallableFunction Type:

Specify something callable with or without async:

```CallableFunction = Union[Callable[..., Awaitable[Any]], Callable]```

Function signature of async function implementation:

```AsyncInnerFunction = Callable[[], Awaitable[Any]]```

"""
# from collections import namedtuple
from typing import Any, Awaitable, Callable, List, NamedTuple, Optional, TypeVar, Union

__all__ = [
    'CallableFunction',
    'AsyncInnerFunction',
    'SUCCESS',
    'FAILURE',
    'ExceptionDecorator',
    'NodeMetadata',
    'node_metadata',
]


CallableFunction = Union[Callable[..., Awaitable[Any]], Callable]
"""Something callable with or without async."""

AsyncInnerFunction = Callable[[], Awaitable[Any]]
"""Function signature of async function implementation."""

SUCCESS = True  # a success call
"""Success constant."""

FAILURE = not SUCCESS  # Well defined falsy...
"""Failure constant."""


class ExceptionDecorator(Exception):
    """ExceptionDecorator exception is a decorator on a real exception.

    This will ensure that ```assert ExceptionDecorator.__bool__ == False```.
    This permit to return exception as a 'FAILURE' status.
    """

    def __init__(self, exception: Exception):
        super().__init__()
        self.exception = exception

    def __bool__(self):
        return False

    def __repr__(self):
        return self.exception.__repr__()

    def __str__(self):
        return self.exception.__str__()


class NodeMetadata(NamedTuple):
    """NodeMetadata is our node definition.

    A NodeMetadata is used to keep information on name, properties name,
    and relations ship name between a hierachical construct of functions.

    This permit us to print or analyze all information of a behaviour tree.

    Attributes:
        name (str): named operation
        properties (List[str]): a list of property name (an int value, ...).
        edges (List[str]): a list of member name which act as edges (a child, ...).

    """

    name: str
    properties: List[str]
    edges: List[str]


T = TypeVar('T', bound=CallableFunction)


def node_metadata(
    name: Optional[str] = None, properties: Optional[List[str]] = None, edges: Optional[List[str]] = None
):
    """'node_metadata' is a function decorator which add meta information about node.

    We add a property on decorated function named '__node_metadata'.

    Args:
        name (Optional[str]): override name of decorated function,
            default is function name left striped with '_'
        properties (Optional[List[str]]): a list of property name ([] as default)
        edges (Optional[List[str]]): a list of edges name
            (["child", "children"] as default)

    Returns:
        the decorator function

    """

    def decorate_function(function: T) -> T:
        function.__node_metadata = NodeMetadata(
            name=name if name else function.__name__.lstrip('_'),
            properties=properties or [],
            edges=edges or ['child', 'children'],
        )
        return function

    return decorate_function



"""Control function definition."""
from typing import Any, List, Optional

from .decorator import is_success
from .definition import FAILURE, AsyncInnerFunction, CallableFunction, ExceptionDecorator, node_metadata

__all__ = ['sequence', 'fallback', 'selector', 'decision', 'repeat_until']


def sequence(children: List[CallableFunction], succes_threshold: int = -1) -> AsyncInnerFunction:
    """Return a function which execute children in sequence.

    succes_threshold parameter generalize traditional sequence/fallback and
    must be in [0, len(children)]. Default value is (-1) means len(children)

    if #success = succes_threshold, return a success

    if #failure = len(children) - succes_threshold, return a failure

    What we can return as value and keep sematic Failure/Success:
     - an array of previous result when success
     - last failure when fail

    Args:
        children (List[CallableFunction]): list of Awaitable
        succes_threshold (int): succes threshold value

    Returns:
        (AsyncInnerFunction): an awaitable function.

    Raises:
        (AssertionError): if succes_threshold is invalid
    """
    succes_threshold = succes_threshold if succes_threshold != -1 else len(children)
    if not (0 <= succes_threshold <= len(children)):
        raise AssertionError('succes_threshold')

    failure_threshold = len(children) - succes_threshold + 1

    @node_metadata(properties=['succes_threshold'])
    async def _sequence():
        success = 0
        failure = 0
        results = []

        for child in children:
            try:
                last_result = await child()
            except Exception as e:
                last_result = ExceptionDecorator(e)

            results.append(last_result)

            if last_result:
                success += 1
                if success == succes_threshold:
                    # last evaluation is a success
                    return results
            else:
                failure += 1
                if failure == failure_threshold:
                    # last evaluation is a failure
                    return last_result
        # should be never reached
        return FAILURE

    return _sequence


def fallback(children: List[CallableFunction]) -> AsyncInnerFunction:
    """Execute tasks in sequence and succeed if one succeed or failed if all failed.

    Often named 'selector', children can be seen as an ordered list
        starting from higthest priority to lowest priority.

    Args:
        children (List[CallableFunction]): list of Awaitable

    Returns:
        (AsyncInnerFunction): an awaitable function.
    """

    return node_metadata(name='fallback')(sequence(children, succes_threshold=min(1, len(children))))


def selector(children: List[CallableFunction]) -> AsyncInnerFunction:
    """Synonym of fallback."""
    return node_metadata(name='selector')(fallback(children))


def decision(
    condition: CallableFunction, success_tree: CallableFunction, failure_tree: Optional[CallableFunction] = None
) -> AsyncInnerFunction:
    """Create a decision node.

    Args:
        condition (CallableFunction): awaitable condition
        success_tree (CallableFunction): awaitable success tree which be
            evaluated if cond is Truthy
        failure_tree (CallableFunction): awaitable failure tree which be
            evaluated if cond is Falsy (None per default)

    Returns:
        (AsyncInnerFunction): an awaitable function.
    """

    @node_metadata(edges=['condition', 'success_tree', 'failure_tree'])
    async def _decision():
        if await condition():
            return await success_tree()
        return await failure_tree() if failure_tree else FAILURE

    return _decision


def repeat_until(condition: CallableFunction, child: CallableFunction) -> AsyncInnerFunction:
    """Repeat child evaluation until condition is truthy.

    Return last child evaluation or FAILURE if no evaluation occurs.

    Args:
        condition (CallableFunction): awaitable condition
        child (CallableFunction): awaitable child

    Returns:
        (AsyncInnerFunction): an awaitable function.
    """

    @node_metadata(edges=['condition', 'child'])
    async def _repeat_until():
        result: Any = FAILURE
        condition_eval = is_success(child=condition)
        while await condition_eval():
            try:
                result = await child()

            except Exception as e:
                result = ExceptionDecorator(e)

        return result

    return _repeat_until



"""Decorator module define all decorator function node."""
from typing import Any

from .definition import FAILURE, SUCCESS, AsyncInnerFunction, CallableFunction, ExceptionDecorator, node_metadata

__all__ = [
    'alias',
    'decorate',
    'always_success',
    'always_failure',
    'is_success',
    'is_failure',
    'inverter',
    'retry',
    'retry_until_success',
    'retry_until_failed',
]


def alias(child: CallableFunction, name: str) -> AsyncInnerFunction:
    """Define an alias on our child.

    Args:
        child (CallableFunction): child function to decorate
        name (str): name of function tree

    Returns:
        (AsyncInnerFunction): an awaitable function.
    """

    @node_metadata(name=name)
    async def _alias():
        return await child()

    return _alias


def decorate(child: CallableFunction, decorator: CallableFunction, **kwargs) -> AsyncInnerFunction:
    """Create a decorator.

    Post process a child with specified decorator function.
    First argument of decorator function must be a child.

    This method implement a simple lazy evaluation.

    Args:
        child (CallableFunction): child function to decorate
        decorator (CallableFunction): awaitable target decorator
        kwargs: optional keyed argument to pass to decorator function

    Returns:
      (AsyncInnerFunction): an awaitable function which
            return decorator evaluation against child.
    """

    @node_metadata(properties=['decorator'])
    async def _decorate():
        return await decorator(await child(), **kwargs)

    return _decorate


def always_success(child: CallableFunction) -> AsyncInnerFunction:
    """Create a node which always return SUCCESS value.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which return child result if is truthy
            else SUCCESS (Any exception will be ignored).

    """

    @node_metadata()
    async def _always_success():
        result: Any = SUCCESS

        try:
            child_result = await child()
            if child_result:
                result = child_result

        except Exception:
            return result

        return result

    return _always_success


def always_failure(child: CallableFunction) -> AsyncInnerFunction:  # -> Awaitable:
    """Produce a function which always return FAILURE value.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which return child result if is falsy
            else FAILURE, or a ControlFlowException if error occurs.

    """

    @node_metadata()
    async def _always_failure():
        result: Any = FAILURE

        try:
            child_result = await child()
            if not child_result:
                result = child_result

        except Exception as e:
            result = ExceptionDecorator(e)

        return result

    return _always_failure


def is_success(child: CallableFunction) -> AsyncInnerFunction:
    """Create a conditional node which test if child success.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which return SUCCESS if child
            return SUCCESS else FAILURE.
            An exception will be evaluated as falsy.
    """

    @node_metadata()
    async def _is_success():
        try:
            return SUCCESS if bool(await child()) else FAILURE
        except Exception as e:
            return ExceptionDecorator(e)

    return _is_success


def is_failure(child: CallableFunction) -> AsyncInnerFunction:
    """Create a conditional node which test if child fail.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which return SUCCESS if child
            return FAILURE else FAILURE.
            An exception will be evaluated as a success.
    """

    @node_metadata()
    async def _is_failure():
        try:
            return SUCCESS if not bool(await child()) else FAILURE
        except Exception:  # pylint: disable=broad-except
            return SUCCESS

    return _is_failure


def inverter(child: CallableFunction) -> AsyncInnerFunction:
    """Invert node status.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which return SUCCESS if child
            return FAILURE else SUCCESS
    """

    @node_metadata()
    async def _inverter():
        return not await child()

    return _inverter


def retry(child: CallableFunction, max_retry: int = 3) -> AsyncInnerFunction:
    """Retry child evaluation at most max_retry time on failure until child succeed.

    Args:
        child (CallableFunction): child function to decorate
        max_retry (int): max retry count (default 3), -1 mean infinite retry

    Returns:
        (AsyncInnerFunction): an awaitable function which retry child evaluation
            at most max_retry time on failure until child succeed.
            If max_retry is reached, returns FAILURE or last exception.
    """
    if not (max_retry > 0 or max_retry == -1):
        raise AssertionError('max_retry')

    @node_metadata(properties=['max_retry'])
    async def _retry():
        retry_count = 0
        infinite_retry_condition = max_retry == -1
        result: Any = FAILURE

        while not result and (infinite_retry_condition or retry_count < max_retry):
            try:
                result = await child()

            except Exception as e:
                # return last failure exception
                if not infinite_retry_condition:  # avoid data allocation if never returned
                    result = ExceptionDecorator(e)

            retry_count += 1

        return result

    return _retry


def retry_until_success(child: CallableFunction) -> AsyncInnerFunction:
    """Retry child until success.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which try to evaluate child
            until it succeed.
    """

    # @node_metadata()
    # async def _retry_until_success():
    #    return await retry(child=child, max_retry=-1)()

    return node_metadata(name='retry_until_success')(retry(child=child, max_retry=-1))


def retry_until_failed(child: CallableFunction) -> AsyncInnerFunction:
    """Retry child until failed.

    Args:
        child (CallableFunction): child function to decorate

    Returns:
        (AsyncInnerFunction): an awaitable function which try to evaluate child
            until it failed.
    """

    return node_metadata(name='retry_until_failed')(retry(child=inverter(child), max_retry=-1))



"""Analyze definition."""
from inspect import getclosurevars
from typing import Any, Callable, List, NamedTuple, Optional, Tuple, no_type_check

from .definition import CallableFunction, NodeMetadata

__all__ = ["analyze", "stringify_analyze", "Node"]


class Node(NamedTuple):
    """Node aggregate node definition implemented with NamedTuple.

    A Node is used to keep information on name, properties, and relations ship
    between a hierachical construct of functions.
    It's like an instance of NodeMetadata.

    Attributes:
        name (str): named operation.
        properties (List[Tuple[str, Any]]): a list of tuple (name, value) for definition.
        edges (List[Tuple[str, List[Any]]]): a list of tuple (name, node list) for
            definition.

    Notes:
        Edges attribut should be edges: ```List[Tuple[str, List['Node']]]```
        But it is impossible for now, see [mypy issues 731](https://github.com/python/mypy/issues/731)
    """

    name: str
    properties: List[Tuple[str, Any]]
    # edges: List[Tuple[str, List['Node']]]
    # https://github.com/python/mypy/issues/731
    edges: List[Tuple[str, List[Any]]]

    def __str__(self):
        return stringify_analyze(target=self)


def _get_analyze_edges(nonlocals) -> Callable:

    return _get_analyze_edges


@no_type_check
def _get_node_metadata(target: CallableFunction) -> NodeMetadata:
    """Returns node metadata instance associated with target."""
    node = target.__node_metadata
    if not isinstance(node, NodeMetadata):
        raise RuntimeError(f'attr __node_metadata of {target} is not a NodeMetadata!')
    return node


# pylint: disable=protected-access
@no_type_check  # it's a shortcut for hasattr ...
def analyze(target: CallableFunction) -> Node:
    """Analyze specified target and return a Node representation.

    Args:
        target (CallableFunction): async function to analyze.

    Returns:
        (Node): a node instance representation of target function
    """

    nonlocals = getclosurevars(target).nonlocals

    def _get_nonlocals_value_for(name):
        return nonlocals[name] if name in nonlocals else None

    def _analyze_property(p):
        """Return a tuple (name, value) or (name, function name) as property."""
        value = _get_nonlocals_value_for(name=p)
        return p, value.__name__ if value and callable(value) else value

    def _analyze_edges(egde_name):
        """Lookup children node from egde_name local var."""
        value = None
        edge = _get_nonlocals_value_for(name=egde_name)
        if edge:
            # it could be a collection of node
            if hasattr(edge, "__iter__"):
                value = list(map(analyze, edge))
            else:  # or a single node
                value = [analyze(edge)]
        return (egde_name, value)

    if hasattr(target, "__node_metadata"):
        node = _get_node_metadata(target=target)
        return Node(
            name=node.name,
            properties=list(map(_analyze_property, node.properties)),
            edges=list(filter(lambda p: p is not None, map(_analyze_edges, node.edges))),
        )

    # simple function
    return Node(
        name=target.__name__.lstrip("_") if hasattr(target, "__name__") else "anonymous",
        properties=list(map(_analyze_property, nonlocals.keys())),
        edges=[],
    )


def stringify_analyze(target: Node, indent: int = 0, label: Optional[str] = None) -> str:
    """Stringify node representation of specified target.

    Args:
        target (CallableFunction): async function to analyze.
        indent (int): level identation (default to zero).
        label (Optional[str]): label of current node (default None).

    Returns:
        (str): a string node representation.
    """
    _ident = '    '
    _space = f'{_ident * indent} '
    result: str = ''
    if label:
        result += f'{_space}--({label})--> {target.name}:\n'
        _space += f"{_ident}{' ' * len(label)}"
    else:
        result += f'{_space}--> {target.name}:\n'

    for k, v in target.properties:
        result += f'{_space}    {k}: {v}\n'

    for _label, children in target.edges:
        if children:
            for child in children:
                result += stringify_analyze(target=child, indent=indent + 1, label=_label)
    return result



"""Declare async btree api."""
from pkg_resources import DistributionNotFound, get_distribution

from .analyze import Node, analyze, stringify_analyze
from .control import decision, fallback, repeat_until, selector, sequence
from .decorator import (
    alias,
    always_failure,
    always_success,
    decorate,
    inverter,
    is_failure,
    is_success,
    retry,
    retry_until_failed,
    retry_until_success,
)
from .definition import (
    FAILURE,
    SUCCESS,
    AsyncInnerFunction,
    CallableFunction,
    ExceptionDecorator,
    NodeMetadata,
    node_metadata,
)
from .leaf import action, condition
from .parallele import parallele
from .utils import afilter, amap, run


__all__ = [
    'Node',
    'analyze',
    'stringify_analyze',
    'decision',
    'fallback',
    'repeat_until',
    'selector',
    'sequence',
    'alias',
    'always_failure',
    'always_success',
    'decorate',
    'inverter',
    'is_failure',
    'is_success',
    'retry',
    'retry_until_failed',
    'retry_until_success',
    'FAILURE',
    'SUCCESS',
    'AsyncInnerFunction',
    'CallableFunction',
    'ExceptionDecorator',
    'NodeMetadata',
    'node_metadata',
    'action',
    'condition',
    'parallele',
    'afilter',
    'amap',
    'run',
]

try:
    __version__ = get_distribution('async-btree').version
except DistributionNotFound:  # pragma: no cover
    __version__ = '(local)'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['async_btree']

package_data = \
{'': ['*']}

extras_require = \
{'curio': ['curio>=1,<2']}

setup_kwargs = {
    'name': 'async-btree',
    'version': '1.0.2',
    'description': 'Async behavior tree',
    'long_description': '# Async Behaviour Tree for Python\n\n\n[![Unix Build Status](https://img.shields.io/travis/geronimo-iia/async-btree/master.svg?label=unix)](https://travis-ci.com/geronimo-iia/async-btree)\n[![Coverage Status](https://img.shields.io/coveralls/geronimo-iia/async-btree/master.svg)](https://coveralls.io/r/geronimo-iia/async-btree)\n[![Codacy Badge](https://api.codacy.com/project/badge/Grade/fe669a02b4aa46b5b1faf619ba2bf382)](https://www.codacy.com/app/geronimo-iia/async-btree?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=geronimo-iia/async-btree&amp;utm_campaign=Badge_Grade)\n[![Scrutinizer Code Quality](https://img.shields.io/scrutinizer/g/geronimo-iia/async-btree.svg)](https://scrutinizer-ci.com/g/geronimo-iia/async-btree/?branch=master)\n[![PyPI Version](https://img.shields.io/pypi/v/async-btree.svg)](https://pypi.org/project/async-btree)\n[![PyPI License](https://img.shields.io/pypi/l/async-btree.svg)](https://pypi.org/project/async-btree)\n\nVersions following [Semantic Versioning](https://semver.org/)\n\nSee [documentation](https://geronimo-iia.github.io/async-btree).\n\n\n## Overview\n\n\n### What\'s a behavior tree ?\n\n> Unlike a Finite State Machine, a Behaviour Tree is a tree of hierarchical nodes that controls the flow of decision and the execution of "tasks" or, as we will call them further, "Actions".\n> -- <cite>[behaviortree](https://www.behaviortree.dev/bt_basics/)</cite>\n\nIf your new (or not) about behavior tree, you could spend some time on this few links:\n\n- [Behavior trees for AI: How they work](https://www.gamasutra.com/blogs/ChrisSimpson/20140717/221339/Behavior_trees_for_AI_How_they_work.php) by Chris Simpson\n- [Introduction to BTs](https://www.behaviortree.dev/bt_basics/)\n\nFew implementation libraries:\n\n- [task_behavior_engine](https://github.com/ToyotaResearchInstitute/task_behavior_engine) A behavior tree based task engine written in Python\n- [pi_trees](https://github.com/pirobot/pi_trees/) a Python/ROS library for implementing Behavior Trees\n- [pr_behavior_tree](https://github.com/personalrobotics/pr_behavior_tree) A simple python behavior tree library based on coroutines\n- [btsk](https://github.com/aigamedev/btsk) Behavior Tree Starter Kit\n- [behave](https://github.com/fuchen/behave) A behavior tree implementation in Python\n\n\n### Why another library so ?\n\n__SIMPLICITY__\n\nWhen you study behavior tree implementation, reactive node, dynamic change, runtime execution, etc ... \nAt a moment you\'re build more or less something that mimic an evaluator \'eval/apply\' or a compilator, with a complex hierachical set of class.\nAll complexity came with internal state management, using tree of blackboard to avoid global variable, multithreading issue, maybe few callback etc ...\nThis break the simplicity and beauty of your initial design.\n\nWhat I find usefull with behavior tree:\n\n- clarity of expression\n- node tree representation\n- possibility to reuse behavior\n- add external measure to dynamicaly change a behavior, a first step on observable pattern...\n\nAs I\'ve used OOP for years (very long time), I will try to avoid class tree and prefer using the power of functionnal programming to obtain what I want: add metadata on a sematic construction, deal with closure, use function in parameters or in return value...\n\nAnd a last reason, more personal, it that i would explore python expressivity.\n\n__SO HOW ?__\n\nIn this module, I purpose you to use the concept of coroutines, and their mecanisms to manage the execution flow.\nBy this way:\n\n- we reuse simple language idiom to manage state, parameter, etc\n- no design constraint on action implementation\n- most of language build block could be reused\n\nYou could build expression like this:\n\n```python\n\nasync def a_func():\n    """A great function"""\n    return "a"\n\nasync def b_decorator(child_value, other=""):\n    """A great decorator..."""\n    return f"b{child_value}{other}"\n\nassert run(decorate(a_func, b_decorator)) == "ba"\n\n```\nThis expression apply ```b_decorator``` on function ```a_func```. \nNote that ```decorate(a_func, b_decorator)``` is not an async function, only action, or condition are async function.\n\n\nFew guidelines of this implementation:\n\n- In order to mimic all NodeStatus (success, failure, running), I replace this by truthy/falsy meaning of evaluation value.\n  A special dedicated exception decorate standard exception in order to give them a Falsy meaning.\n- Blackboard pattern, act as a manager of context variable for behavior tree.\n  With python 3, please... simply use [contextvars](https://docs.python.org/3/library/contextvars.html) !\n- In order to be able to build a sematic tree, I\'ve introduce a metadata tuple added on function implementation.\n\nThe rest is just implementation details..\n\n\n\nA little note:\n\n> You should not use this until you\'re ready to think about what you\'re doing :)\n\n\n### Note about \'async\' framework\n\nAs we use async function as underlaying mechanism to manage the execution flow, the standard library asyncio is pretty fine.\nBut, (always a but somewhere isn\'t it...), you should read this [amazing blog post}(https://vorpus.org/blog/some-thoughts-on-asynchronous-api-design-in-a-post-asyncawait-world/) by Nathaniel J. Smith.\nAnd next study [curio](https://github.com/dabeaz/curio) framework in deep.\n\nAs curio say:\n> Don\'t Use Curio if You\'re Allergic to Curio\n\nPersonaly, after few time of testing and reading curio code, I\'m pretty addict.\n\n## Installation\n\nInstall this library directly into an activated virtual environment:\n\n```text\n$ pip install async-btree\n```\n\nor add it to your [Poetry](https://poetry.eustace.io/) project:\n\n```text\n$ poetry add async-btree\n```\n\n## Usage\n\nAfter installation, the package can imported:\n\n```text\n$ python\n>>> import async_btree\n>>> async_btree.__version__\n```\n\nSee [API Reference documentation](https://geronimo-iia.github.io/async-btree).\n\n\nWith this framework, you didn\'t find any configuration file, no Xml, no json, no yaml.\n\nThe main reason (oriented and personal point of view) is that you did not need to introduce an extra level of abtraction \nto declare a composition of functions. I think it\'s true for most of main use case (except using an editor to wrote behaviour tree for example).\n\nSo "If you wrote your function with python, wrote composition in python"... \n_(remember that you did not need XML to do SQL, just write good sql...)_\n\n\nSo, the goal is to:\n - define your business function wich implements actions or conditions, with all test case that you wish/need\n - compose them using those provided by this framework like ```sequence```, ```selector```, ...\n - use them as it is or create a well define python module to reuse them\n\n\nWanna style have an abtract tree of our behaviour tree ?\n\nFunctions from async-btree build an abstract tree for you. \nIf you lookup in code, you should see an annotation "node_metadata" on internal implementation. \nThis decorator add basic information like function name, parameters, and childreen relation ship.\n\nThis abstract tree can be retreived and stringified with ```analyze``` and ```stringify_analyze```.\nHere the profile:\n\n```python\n  def analyze(target: CallableFunction) -> Node: # here we have our "abtract tree code"\n    ...\n```\n\nFor example:\n\n```python\n\n# your behaviour tree, or a sub tree:\nmy_func = alias(child=repeat_until(child=action(hello), condition=success_until_zero), name="btree_1")\n\n# retrieve meta information and build a Node tree\nabstract_tree_tree_1 = analyze(my_func) \n\n# output the tree:\nprint(stringify_analyze(abstract_tree_tree_1))\n```\n\nThis should print:\n\n```text\n --> btree_1:\n     --(child)--> repeat_until:\n         --(condition)--> success_until_zero:\n         --(child)--> action:\n                      target: hello\n```',
    'author': 'Jerome Guibert',
    'author_email': 'jguibert@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/async_btree',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Leaf definition."""

from .decorator import is_success
from .definition import AsyncInnerFunction, CallableFunction, ExceptionDecorator, node_metadata

__all__ = ['action', 'condition']


def action(target: CallableFunction, **kwargs) -> AsyncInnerFunction:
    """Declare an action leaf.

    Action is an awaitable closure of specified function.

    Args:
        target (CallableFunction): awaitable function
        kwargs: optional kwargs argument to pass on target function

    Returns:
        (AsyncInnerFunction): an awaitable function.
    """

    @node_metadata(properties=['target'])
    async def _action():
        try:
            return await target(**kwargs)
        except Exception as e:
            return ExceptionDecorator(exception=e)

    return _action


def condition(target: CallableFunction, **kwargs) -> AsyncInnerFunction:
    """Declare a condition leaf.

    Condition is an awaitable closure of specified function.

    Args:
        target (CallableFunction):  awaitable function which be evaluated as True/False.
        kwargs: optional kwargs argument to pass on target function

    Returns:
        (AsyncInnerFunction): an awaitable function.
    """

    return node_metadata(name='condition', properties=['target'])(is_success(action(target=target, **kwargs)))



"""Utility function."""
from typing import Any, AsyncGenerator, AsyncIterable, Awaitable, Callable, Iterable, TypeVar, Union

__all__ = ['amap', 'afilter', 'run']

T = TypeVar('T')


async def amap(
    corofunc: Callable[[Any], Awaitable[T]], iterable: Union[AsyncIterable, Iterable]
) -> AsyncGenerator[T, None]:
    """Map an async function onto an iterable or an async iterable.

    This simplify writing of mapping a function on something iterable
    between 'async for ...' and 'for...' .

    Args:
        corofunc (Callable[[Any], Awaitable[T]]): coroutine function
        iterable (Union[AsyncIterable, Iterable]): iterable or async iterable collection
            which will be applied.

    Returns:
        AsyncGenerator[T]: an async iterator of corofunc(item)

    Example:
        ```[i async for i in amap(inc, afilter(even, [0, 1, 2, 3, 4]))]```

    """
    if isinstance(iterable, AsyncIterable):  # if hasattr(iterable, '__aiter__'):
        async for item in iterable:
            yield await corofunc(item)
    else:
        for item in iterable:
            yield await corofunc(item)


async def afilter(
    corofunc: Callable[[Any], Awaitable[bool]], iterable: Union[AsyncIterable, Iterable]
) -> AsyncGenerator[T, None]:
    """Filter an iterable or an async iterable with an async function.

    This simplify writing of filtering by a function on something iterable
    between 'async for ...' and 'for...' .

    Args:
        corofunc (Callable[[Any], Awaitable[bool]]): filter async function
        iterable (Union[AsyncIterable, Iterable]): iterable or async iterable collection
            which will be applied.

    Returns:
        (AsyncGenerator[T]): an async iterator of item which satisfy corofunc(item) == True

    Example:
        ```[i async for i in amap(inc, afilter(even, [0, 1, 2, 3, 4]))]```

    """
    if isinstance(iterable, AsyncIterable):  # if hasattr(iterable, '__aiter__'):
        async for item in iterable:
            if await corofunc(item):
                yield item
    else:
        for item in iterable:
            if await corofunc(item):
                yield item


try:
    import curio  # noqa: F401
    from contextvars import copy_context

    def run(kernel, target, *args):
        """Curio run with independent contextvars.

        This mimic asyncio framework behaviour.

        ```
        copy_context().run(kernel.run, target, *args)
        ```

        """
        return copy_context().run(kernel.run, target, *args)


except Exception:  # pragma: no cover

    def run(kernel, target, *args):
        raise RuntimeError('curio not installed!')
